using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using Battlehub.RTSaveLoad.PersistentObjects.UI;
using SimpleJSON;
using UnityEngine;

namespace CheesyFX
{
    public class JiggleHandler : MonoBehaviour
    {
        public Rigidbody rb;
        private RBDynamics rbDynamic;

        private bool[] isJiggling = new bool[4];
        private float[] amp = new float[4];
        private float[] phase = new float[4];
        private float[] sin = new float[4];
        public Dictionary<DAZMorph, float>[] morphs = new Dictionary<DAZMorph, float>[4];

        public JSONStorableVector4 strengthJSON = new JSONStorableVector4(
            "strength",
            new Vector4(.02f, .02f, .02f, .2f),
            new Vector4(),
            new Vector4(1f,1f,1f, 1f));
        
        public JSONStorableVector4 maxMorph = new JSONStorableVector4(
            "maxMorph",
            new Vector4(1f, 1f, 1f, 1f),
            new Vector4(),
            new Vector4(3f,3f,3f, 3f));
        
        public JSONStorableVector4 damping = new JSONStorableVector4(
            "damping",
            new Vector4(.05f, .05f, .05f,05f),
            new Vector4(),
            new Vector4(1f,1f,1f,1f));
        
        public JSONStorableVector4 frequency = new JSONStorableVector4(
            "frequency",
            new Vector4(15f, 15, 15f,15f),
            new Vector4(),
            new Vector4(25f,25f,25f,25f));
        
        public JSONStorableVector4 minAccel = new JSONStorableVector4(
            "minAccel",
            new Vector4(.02f, .02f, .02f,.02f),
            new Vector4(),
            new Vector4(.1f,.1f,.1f,.1f));
        
        public JSONStorableVector4 maxAccel = new JSONStorableVector4(
            "maxAccel",
            new Vector4(1f, 1f, 1f,1f),
            new Vector4(),
            new Vector4(5f,5f,5f,5f));
        
        public JSONStorableVector3 topMidBottomScale = new JSONStorableVector3(
            "Scale",
            new Vector4(1f, 1f, 1f),
            new Vector4(),
            new Vector4(2f,2f,2f));

        
        private const float TwoPi = 2f*Mathf.PI;
        private const float PhaseToId = 360f/(2f*Mathf.PI);
        private float _directionFactor = 1f;

        public void Init(Rigidbody rigidbody)
        {
            rb = rigidbody;
            // strengthJSON.setCallbackFunction += val => $"{rb.name}: {val}".Print();
            switch ( rb.name)
            {
                case "lThigh":
                {
                    strengthJSON.defaultVal = new Vector4(.03f, .02f, .01f,.025f);
                    damping.defaultVal = new Vector4(.07f, .08f, .1f, .08f);
                    frequency.defaultVal = new Vector4(17f, 18f, 20f, 17f);
                    morphs[0] = new Dictionary<DAZMorph, float>
                    {
                        {ShakeIt.GetMorphByUid("SI_lThighTop_X"), 1f},
                        {ShakeIt.GetMorphByUid("SI_lThighMid_X"), 1f},
                        {ShakeIt.GetMorphByUid("SI_lThighLow_X"), .1f}
                    };

                    morphs[1] = new Dictionary<DAZMorph, float>
                    {
                        {ShakeIt.GetMorphByUid("SI_lThighTop_Y"), .5f},
                        {ShakeIt.GetMorphByUid("SI_lThighMid_Y"), .5f},
                        {ShakeIt.GetMorphByUid("SI_lThighLow_Y"), 1f}
                    };
                    morphs[2] = new Dictionary<DAZMorph, float>
                    {
                        {ShakeIt.GetMorphByUid("SI_lThighTop_Z"), 1f},
                        {ShakeIt.GetMorphByUid("SI_lThighMid_Z"), 1f},
                        {ShakeIt.GetMorphByUid("SI_lThighLow_Z"), 1f}
                    };
                    morphs[3] = new Dictionary<DAZMorph, float>
                    {
                        {ShakeIt.GetMorphByUid("SI_lThighTop_R"), 1f},
                        {ShakeIt.GetMorphByUid("SI_lThighMid_R"), 1f},
                        {ShakeIt.GetMorphByUid("SI_lThighLow_R"), 1f}
                    };
                    break;
                }
                case "rThigh":
                {
                    strengthJSON.defaultVal = new Vector4(.03f, .02f, .01f, .025f);
                    maxMorph.defaultVal = new Vector4(1f, 1f, 1f, 1f);
                    damping.defaultVal = new Vector4(.07f, .08f, .1f, .08f);
                    frequency.defaultVal = new Vector4(17f, 18f, 20f,17f);
                    morphs[0] = new Dictionary<DAZMorph, float>
                    {
                        {ShakeIt.GetMorphByUid("SI_rThighTop_X"), 1f},
                        {ShakeIt.GetMorphByUid("SI_rThighMid_X"), 1f},
                        {ShakeIt.GetMorphByUid("SI_rThighLow_X"), 1f}
                    };
                    morphs[1] = new Dictionary<DAZMorph, float>
                    {
                        {ShakeIt.GetMorphByUid("SI_rThighTop_Y"), .75f},
                        {ShakeIt.GetMorphByUid("SI_rThighMid_Y"), .75f},
                        {ShakeIt.GetMorphByUid("SI_rThighLow_Y"), 1f}
                    };
                    morphs[2] = new Dictionary<DAZMorph, float>
                    {
                        {ShakeIt.GetMorphByUid("SI_rThighTop_Z"), 1f},
                        {ShakeIt.GetMorphByUid("SI_rThighMid_Z"), 1f},
                        {ShakeIt.GetMorphByUid("SI_rThighLow_Z"), 1f}
                    };
                    morphs[3] = new Dictionary<DAZMorph, float>
                    {
                        {ShakeIt.GetMorphByUid("SI_rThighTop_R"), 1f},
                        {ShakeIt.GetMorphByUid("SI_rThighMid_R"), 1f},
                        {ShakeIt.GetMorphByUid("SI_rThighLow_R"), 1f}
                    };
                    break;
                }
                case "lShin":
                {
                    _directionFactor = -1f;
                    strengthJSON.defaultVal = new Vector4(0f, 0f, 0f, .02f);
                    frequency.defaultVal = new Vector4(0f,0f,0f,16f);
                    damping.defaultVal = new Vector4(.1f, .065f, .05f,.08f);
                    morphs[3] = new Dictionary<DAZMorph, float>
                    {
                        {ShakeIt.GetMorphByUid("SI_lCalf_R"), 1f}
                    };
                    break;
                }
                case "rShin":
                {
                    strengthJSON.defaultVal = new Vector4(.0f, 0f, 0f, .02f);
                    frequency.defaultVal = new Vector4(0f,0f,0f,16f);
                    damping.defaultVal = new Vector4(.1f, .065f, .05f,.08f);
                    
                    morphs[3] = new Dictionary<DAZMorph, float>
                    {
                        {ShakeIt.GetMorphByUid("SI_rCalf_R"), 1f}
                    };
                    break;
                }
                case "lShldr":
                {
                    _directionFactor = -1f;
                    strengthJSON.defaultVal = new Vector4(.015f, .01f, .015f, .015f);
                    damping.defaultVal = new Vector4(.07f, .07f, .07f, .07f);
                    minAccel.defaultVal = new Vector4(.05f, .05f, .05f, .05f);
                    morphs[3] = new Dictionary<DAZMorph, float>
                    {
                        {ShakeIt.GetMorphByUid("SI_lArmTop_R"), .75f},
                        {ShakeIt.GetMorphByUid("SI_lArmMid_R"), 1f},
                        {ShakeIt.GetMorphByUid("SI_lArmLow_R"), .75f}
                    };

                    morphs[1] = new Dictionary<DAZMorph, float>
                    {
                        {ShakeIt.GetMorphByUid("SI_lArmTop_Y"), .75f},
                        {ShakeIt.GetMorphByUid("SI_lArmMid_Y"), 1f},
                        {ShakeIt.GetMorphByUid("SI_lArmLow_Y"), .75f}
                    };
                    break;
                }
                case "rShldr":
                {
                    _directionFactor = 1f;
                    strengthJSON.defaultVal = new Vector4(.015f, .01f, .015f, .015f);
                    damping.defaultVal = new Vector4(.07f, .07f, .07f, .07f);
                    minAccel.defaultVal = new Vector4(.05f, .05f, .05f, .05f);
                    morphs[3] = new Dictionary<DAZMorph, float>
                    {
                        {ShakeIt.GetMorphByUid("SI_rArmTop_R"), .75f},
                        {ShakeIt.GetMorphByUid("SI_rArmMid_R"), 1f},
                        {ShakeIt.GetMorphByUid("SI_rArmLow_R"), .75f}
                    };

                    morphs[1] = new Dictionary<DAZMorph, float>
                    {
                        {ShakeIt.GetMorphByUid("SI_rArmTop_Y"), .75f},
                        {ShakeIt.GetMorphByUid("SI_rArmMid_Y"), 1f},
                        {ShakeIt.GetMorphByUid("SI_rArmLow_Y"), .75f}
                    };
                    break;
                }
                case "abdomen":
                    strengthJSON.defaultVal = new Vector4(.02f, .02f, .02f);
                    frequency.defaultVal = new Vector4(20f, 20f, 20f);
                    damping.defaultVal = new Vector4(.08f, .08f, .11f);
                    maxAccel.defaultVal = new Vector4(1.5f, 1.5f, 1.5f, 1.5f);
                    morphs[0] = new Dictionary<DAZMorph, float>
                    {
                        {ShakeIt.GetMorphByUid("SI_AbsLow_X"), 1f},
                        {ShakeIt.GetMorphByUid("SI_AbsSidesLow_X"), .75f},
                        {ShakeIt.GetMorphByUid("SI_AbsMid_X"), .75f},
                        {ShakeIt.GetMorphByUid("SI_AbsSidesMid_X"), .75f}
                    };

                    morphs[1] = new Dictionary<DAZMorph, float>
                    {
                        {ShakeIt.GetMorphByUid("SI_AbsLow_Y"), 1f},
                        {ShakeIt.GetMorphByUid("SI_AbsMid_Y"), .5f}
                    };
                    
                    morphs[2] = new Dictionary<DAZMorph, float>
                    {
                        {ShakeIt.GetMorphByUid("SI_AbsLow_Z"), .5f},
                        {ShakeIt.GetMorphByUid("SI_AbsMid_Z"), .5f}
                    };
                    break;
                case "abdomen2":
                    strengthJSON.defaultVal = new Vector4(.02f, .015f, .015f);
                    frequency.defaultVal = new Vector4(20f, 20f, 20f);
                    damping.defaultVal = new Vector4(.08f, .08f, .08f);
                    maxAccel.defaultVal = new Vector4(1.5f, 1.5f, 1.5f, 1.5f);
                    morphs[0] = new Dictionary<DAZMorph, float>
                    {
                        {ShakeIt.GetMorphByUid("SI_AbsTop_X"), .75f},
                        {ShakeIt.GetMorphByUid("SI_AbsSidesTop_X"), .5f},
                    };

                    morphs[1] = new Dictionary<DAZMorph, float>
                    {
                        {ShakeIt.GetMorphByUid("SI_AbsTop_Y"), .5f},
                        
                    };
                    
                    morphs[2] = new Dictionary<DAZMorph, float>
                    {
                        {ShakeIt.GetMorphByUid("SI_AbsTop_Z"), .25f},
                        
                    };
                    break;

            } 
            // morphs = new List<DAZMorph>{rotUp, rotMid, rotLow, transUp, transMid, transLow};
            strengthJSON.val = strengthJSON.defaultVal;
            maxMorph.val = maxMorph.defaultVal;
            frequency.val = frequency.defaultVal;
            damping.val = damping.defaultVal;
            maxAccel.val = maxAccel.defaultVal;
            minAccel.val = minAccel.defaultVal;
            Reset();
            rbDynamic = rb.GetComponent<RBDynamics>();
        }

        public void Reset()
        {
            foreach (var morphSet in morphs)
            {
                if(morphSet == null) continue;
                foreach (var dazMorph in morphSet.Keys.ToList()) dazMorph.Reset();
            }
        }

        public void OnDisable()
        {
            Reset();
            // rb.name.Print();
        }

        public void OnDestroy()
        {
            Reset();
        }

        public void FixedUpdate()
        {
            float[] mag = new float[4];
            Transform transform = rb.transform;
            Vector3 acc = rbDynamic.acceleration;
            acc = transform.InverseTransformDirection(acc);
            Vector3 localAngularVel = transform.InverseTransformDirection(rb.angularVelocity);
            for(int i = 0; i<4; i++)
            {
                if(morphs[i] == null) continue;
                switch (i)
                {
                    case 0:
                    {
                        mag[i] = acc.x * _directionFactor;
                        break;
                    }
                    case 1:
                    {
                        mag[i] = acc.y + localAngularVel.x;
                        break;
                    }
                    case 2:
                    {
                        mag[i] = acc.z;
                        break;
                    }
                    case 3:
                    {
                        mag[i] = Mathf.Sqrt(acc.x * acc.x + acc.z * acc.z);
                        break;
                    }
                }
                // if(Mathf.Abs(mag[i]) < minAccel.val[i]) continue;
                if (Mathf.Abs(mag[i]) > maxAccel.val[i]) mag[i] = maxAccel.val[i] * Mathf.Sign(mag[i]);
                Jiggle(i, mag[i]);
            }
        }

        private void Jiggle(int dir, float strength)
        {
            
            float a = amp[dir];
            
            if (Math.Abs(strength) > minAccel.val[dir])
            {
                strength *= strengthJSON.val[dir];
                if (!isJiggling[dir])
                {
                    a = Clamp(strength, dir);
                    isJiggling[dir] = true;
                }
                else
                {
                    // float newAmp = Clamp(a + Mathf.Abs(strength)*Mathf.Sign(a), dir);
                    float newAmp = Clamp(a + strength, dir);
                    a = newAmp;
                }
            }
            if (Mathf.Abs(a) > .005f)
            {
                a = Mathf.Lerp(a, 0f, 60f*damping.val[dir]*Time.fixedUnscaledDeltaTime);
                float p = phase[dir];
                p += frequency.val[dir] * Time.fixedUnscaledDeltaTime;
                if (p > TwoPi) p = 0f;
                float _sin = sin[dir] = a * ShakeIt.sin[(int)(p * PhaseToId)];
                foreach (var item in morphs[dir]) item.Key.morphValue = item.Value * _sin;
                
                amp[dir] = a;
                phase[dir] = p;
            }
            else if (isJiggling[dir])
            {
                phase[dir] = 0f;
                isJiggling[dir] = false;
            }
        }
        
        private float Clamp(float val, int dir)
        {
            var mVal = maxMorph.val[dir];
            if (val > mVal) return mVal;
            if (val < -mVal) return -mVal;
            return val;
        }

        public bool Store(JSONClass jc, bool forceStore = true)
        {
            JSONClass tempJC = new JSONClass();
            bool needsStore = strengthJSON.Store(tempJC);
            if (maxMorph.Store(tempJC, forceStore)) needsStore = true;
            if (frequency.Store(tempJC, forceStore)) needsStore = true;
            if (damping.Store(tempJC, forceStore)) needsStore = true;
            if (minAccel.Store(tempJC, forceStore)) needsStore = true;
            if (maxAccel.Store(tempJC, forceStore)) needsStore = true;

            if (needsStore)
            {
                jc[rb.name] = tempJC;
                return true;
            }
            return false;
        }

        public void Load(JSONClass jc)
        {
            if (jc.HasKey(rb.name))
            {
                jc = jc[rb.name].AsObject;
                maxMorph.Load(jc);
                strengthJSON.Load(jc);
                frequency.Load(jc);
                damping.Load(jc);
                minAccel.Load(jc);
            }
        }
    }
}