using System;
using System.Collections.Generic;
using System.Linq;
using MacGruber;
using MVR.FileManagementSecure;
using SimpleJSON;
using UnityEngine;
using UnityEngine.Assertions.Must;
using UnityEngine.UI;


namespace CheesyFX
{
    public class ShakeIt : MVRScript
    {
        private static string morphPath = "Custom/Atom/Person/Morphs/female/CheesyFX/ShakeIt/";
        private const string saveDir = "Saves/PluginData/CheesyFX/ShakeIt/";
        private static string factoryDefaults;

        private List<JiggleHandler> selectedHandlers;

        private static string[] jiggleZones = { "Thighs", "Calves", "Arms", "Abdomen", "Abdomen2" };
        private static Dictionary<int, List<JiggleHandler>> jiggleHandlersByRegion = new Dictionary<int, List<JiggleHandler>>();
        private List<JiggleHandler> jiggleHandlers = new List<JiggleHandler>();

        public static Dictionary<int, float> sin = new Dictionary<int, float>();

        private UIDynamicV4Slider maxMorph;
        private UIDynamicV4Slider strengthUid;
        private UIDynamicV4Slider frequencyUid;
        private UIDynamicV4Slider dampingUid;
        private UIDynamicV4Slider thresholdUid;
        private UIDynamicV4Slider maxAccelUid;
        private string lastSymChoice;

        private static GenerateDAZMorphsControlUI morphUI;
        public JSONStorableUrl loadURL;
        private List<RBDynamics> rbDynamics = new List<RBDynamics>();
        private bool needsStore;

        private bool develop = false;

        public override void Init()
        {
            DAZCharacterSelector dcs = containingAtom.GetComponentInChildren<DAZCharacterSelector>();
            morphUI = dcs.morphsControlUI;
            if (!morphUI)
            {
                "MorphUI not ready".Print();
                return;
            }
            FileManagerSecure.CreateDirectory(saveDir);
            factoryDefaults = "Custom/Scripts/CheesyFX/ShakeIt/FactoryDefaults.si";
            var packageUid = CheesyUtils.GetLatestPackageUid("CheesyFX.ShakeIt")+":/";
            if (!develop)
            {
                if (packageUid != null)
                {
                    factoryDefaults = packageUid + factoryDefaults;
                    morphPath = packageUid + morphPath;
                }
                else return;
            }
            
            loadURL = new JSONStorableUrl("loadURL", "", UILoadJSON, "jpg", true);
            loadURL.hideExtension = true;
            loadURL.setCallbackFunction -= UILoadJSON;
            loadURL.allowFullComputerBrowse = false;
            loadURL.allowBrowseAboveSuggestedPath = true;
            loadURL.SetFilePath(saveDir);
            loadURL.setCallbackFunction += UILoadJSON;
            List<string> rbNames = new List<string>
            {
                "lThigh", "rThigh", "lShin", "rShin", "rShldr", "lShldr", "abdomen", "abdomen2"
            };
            for (int i=0; i<rbNames.Count; i++)
            {
                var rbName = rbNames[i];
                var rb = containingAtom.forceReceivers.First(x => x.name == rbName).GetComponent<Rigidbody>();
                rbDynamics.Add(rb.AddRbDynamic());
                var handler = gameObject.AddComponent<JiggleHandler>();
                handler.Init(rb);
                jiggleHandlers.Add(handler);
                if(i == 0) jiggleHandlersByRegion[0] = new List<JiggleHandler>{handler};
                else if(i == 1) jiggleHandlersByRegion[0].Add(handler);
                else if(i == 2) jiggleHandlersByRegion[1] = new List<JiggleHandler>{handler};
                else if(i == 3) jiggleHandlersByRegion[1].Add(handler);
                else if(i == 4) jiggleHandlersByRegion[2] = new List<JiggleHandler>{handler};
                else if(i == 5) jiggleHandlersByRegion[2].Add(handler);
                else if(i == 6) jiggleHandlersByRegion[3] = new List<JiggleHandler>{handler};
                else if(i == 7) jiggleHandlersByRegion[4] = new List<JiggleHandler>{handler};
            }
            const float twoPi = 2f * Mathf.PI;
            for (int i = 0; i < 360; i++)
            {
                sin[i] = Mathf.Sin(i/360f*twoPi);
            }
            UIManager.OnInitUI(CreateUIElement, leftUIElements, rightUIElements);
            Utils.OnInitUI(CreateUIElement);
            CreateUI();
            ApplyLatestMatchingPreset();

        }
        
        public void OnEnable()
        {
            for (int i=0; i<jiggleHandlers.Count; i++) jiggleHandlers[i].enabled = true;
        }
        
        public void OnDisable()
        {
            for (int i=0; i<jiggleHandlers.Count; i++) jiggleHandlers[i].enabled = false;
        }

        private void OnDestroy()
        {
            for (int i = 0; i < rbDynamics.Count; i++)
            {
                rbDynamics[i].uses--;
                if (rbDynamics[i].uses == 0)
                {
                    Destroy(rbDynamics[i]);
                }
            }
        }

        private void CreateUI()
        {
            var twinButton = Utils.SetupTwinButton(this,
                "Save preset", UISaveJSONDialog,
                "Load preset", delegate {}, false);
            loadURL.RegisterFileBrowseButton(twinButton.buttonRight);
            
            Utils.SetupTwinButton(this,
                "UserDefaults", () => ApplyFromUrl(saveDir+"UserDefaults.si"),
                "FactoryDefaults", () => ApplyFromUrl(factoryDefaults), true);
            
            var tabbar = UIManager.CreateTabBar(jiggleZones, OnRegionChanged, script:this, invokeDefault:false);
            strengthUid = UIManager.CreateV4Slider(script: this);
            frequencyUid = UIManager.CreateV4Slider(script: this);
            dampingUid = UIManager.CreateV4Slider(script: this);
            thresholdUid = UIManager.CreateV4Slider(script: this);
            maxAccelUid = UIManager.CreateV4Slider(script: this);
            maxMorph = UIManager.CreateV4Slider(script: this);
            tabbar.buttons[0].onClick.Invoke();
        }

        private void OnRegionChanged(int id)
        {
            selectedHandlers = jiggleHandlersByRegion[id];
            
            maxMorph.RegisterVectors(selectedHandlers.Select(x => x.maxMorph).ToList());
            strengthUid.RegisterVectors(selectedHandlers.Select(x => x.strengthJSON).ToList());
            frequencyUid.RegisterVectors(selectedHandlers.Select(x => x.frequency).ToList());
            dampingUid.RegisterVectors(selectedHandlers.Select(x => x.damping).ToList());
            thresholdUid.RegisterVectors(selectedHandlers.Select(x => x.minAccel).ToList());
            maxAccelUid.RegisterVectors(selectedHandlers.Select(x => x.maxAccel).ToList());
        }

        // public static DAZMorph GetMorphByDisplayName(string name)
        // {
        //     DAZMorph morph;
        //     morph = morphUI.GetMorphByDisplayName(name);
        //     morph.uid.Print();
        //     // if(morph == null) $"Morph '{name}' not found!".Print();
        //     return morph;
        // }
        
        public static DAZMorph GetMorphByUid(string name)
        {
            DAZMorph morph;
            string uid = morphPath + name +".vmi";
            morph = morphUI.GetMorphByUid(uid);
            // if(morph != null) uid.Print();
            if(morph == null) $"Morph '{uid}' not found!".Print();
            return morph;
        }
        
        public JSONClass MyGetJSON()
        {
            JSONClass jc = new JSONClass();
            needsStore = false;
            for (int i = 0; i < jiggleHandlers.Count; i++)
            {
                if (jiggleHandlers[i].Store(jc) && !needsStore) needsStore = true;
            }
            return jc;
        }
        
        private void ApplyFromUrl(string path)
        {
            if(!FileManagerSecure.FileExists(path)) return;
            path = FileManagerSecure.NormalizePath(path);
            {
                JSONClass jc = LoadJSON(path).AsObject;
                ApplyJSON(jc);
            }
        }

        public void ApplyJSON(JSONClass jc)
        {
            for (int i = 0; i < jiggleHandlers.Count; i++)
            {
                jiggleHandlers[i].Load(jc);
            }
        }
        
        public void UISaveJSONDialog(){
            SuperController sc = SuperController.singleton;
            sc.GetMediaPathDialog(UISaveJSON, "jpg", saveDir, false, true, false, null, true, null, false, false);
            sc.mediaFileBrowserUI.SetTextEntry(true);
            if(sc.mediaFileBrowserUI.fileEntryField != null){
                string filename = GetSuggestedPresetName();
                sc.mediaFileBrowserUI.fileEntryField.text = filename;
                sc.mediaFileBrowserUI.ActivateFileNameField();
            }
            
        }
        
        public void UILoadJSON(string url){
            JSONClass jc = LoadJSON(url.Replace(".jpg",".si")).AsObject;
            if(jc != null) ApplyJSON(jc);
        }
        
        public void UISaveJSON(string path)
        {
            if (string.IsNullOrEmpty(path))
                return;
            string path1 = path.Replace('\\', '/')+".si";
            JSONClass jc = MyGetJSON();
            SaveJSON(jc, path1);
            SuperController.singleton.DoSaveScreenshot(path+".jpg");
        }

        private void ApplyLatestMatchingPreset()
        {
            List<string> matchingPresetPaths = new List<string>();
            string atomName = containingAtom.name;
            string[] paths = FileManagerSecure.GetFiles(saveDir);
            for (int i = 0; i < paths.Length; i++)
            {
                string path = paths[i];
                if (!path.EndsWith(".si")) continue;
                string presetName = path.Replace(saveDir+"\\","");
                
                if(presetName.StartsWith(atomName)) matchingPresetPaths.Add(presetName);
                else if (presetName.StartsWith("UserDefaults")) matchingPresetPaths.Insert(0, presetName);
            }

            if (matchingPresetPaths.Count > 0)
            {
                string prestName = matchingPresetPaths.Last();
                ApplyFromUrl(saveDir+prestName);
                $"Successfully applied '{saveDir+prestName}'.".Print();
            }
            // else $"No matching preset for '{atomName}' found.".Print();
        }

        private string GetSuggestedPresetName()
        {
            List<string> presetNames = new List<string>();
            string atomName = containingAtom.name;
            string[] paths = FileManagerSecure.GetFiles(saveDir);
            for (int i = 0; i < paths.Length; i++)
            {
                string path = paths[i];
                if (!path.EndsWith(".si")) continue;
                string preset = path.Replace(saveDir+"\\","");
                preset = preset.Remove(preset.Length - 3);
                presetNames.Add(preset);
            }
            string suggested;
            for (int i = 1; i < 10; i++)
            {
                suggested = atomName+ "." + i.ToString("00");
                if (!presetNames.Contains(suggested)) return suggested;
            }
            return atomName;
        }
        
        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
            if(needsStore) jc["ShakeIt"] = MyGetJSON();
            return jc;
        }
        
        public override void LateRestoreFromJSON(
            JSONClass jc,
            bool restorePhysical,
            bool restoreAppearance,
            bool setMissingToDefault
        )
        {
            base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
            if (restorePhysical && restoreAppearance) // StoreType.Full
            {
                if (jc.HasKey("ShakeIt")){
                    JSONClass jcShakeIt = jc["ShakeIt"].AsObject;
                    ApplyJSON(jcShakeIt);
                }
                
            }
        }
    }
}