﻿using System;
using System.Linq;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.Events;
using UnityEngine.Scripting;
using SimpleJSON;
using MeshVR;
using MVR.FileManagementSecure;
using MacGruber;
//using System.Runtime.CompilerServices;
//using System.Core.dll;

namespace CheesyFX
{
	public static class VectorExt
	{
		public static Vector3 ProjectY(this Vector3 v)
		{
			return new Vector3(v.x, 0f, v.z);
		}

		public static Vector3 MirrorY(this Vector3 v)
		{
			return new Vector3(v.x, -v.y, v.z);
		}

		public static float RadialDistance(this Vector3 v, Vector3 w)
		{
			return (ProjectY(v) - ProjectY(w)).magnitude;
		}

		public static float RadialSqrDistance(this Vector3 v, Vector3 w)
		{
			return (ProjectY(v) - ProjectY(w)).sqrMagnitude;
		}
		
		public static Vector3 Scale(this Vector3 v, Vector3 w)
		{
			return new Vector3(v.x * w.x, v.y*w.y, v.z*w.z);
		}

		public static JSONArray ToJA(this Vector3 v)
		{
			JSONArray ja = new JSONArray();
			for (int i = 0; i < 3; i++)
				ja[i].AsFloat = v[i];
			return ja;
		}
		
		public static JSONArray ToJA(this Vector4 v)
		{
			JSONArray ja = new JSONArray();
			for (int i = 0; i < 4; i++)
				ja[i].AsFloat = v[i];
			return ja;
		}

		public static Vector2 Scale(this Vector2 v, Vector2 w)
		{
			return new Vector2(v.x * w.x, v.y * w.y);
		}
		
		public static Vector2 ScaleX(this Vector2 v, float scale)
		{
			return new Vector2(v.x * scale, v.y);
		}
		
		public static Vector2 ScaleY(this Vector2 v, float scale)
		{
			return new Vector2(v.x, v.y * scale);
		}
		
		public static Vector2 SetX(this Vector2 v, float val)
		{
			return new Vector2(val, v.y);
		}
		
		public static Vector2 SetY(this Vector2 v, float val)
		{
			return new Vector2(v.x, val);
		}
		
		public static Vector3 ToV3(this Vector4 v)
		{
			return new Vector3(v.x, v.y, v.z);
		}
		
		public static Vector4 ToV4(this Vector3 v, float w = 0f)
		{
			return new Vector4(v.x, v.y, v.z, w);
		}

	}
	
	public static class TransformExt
	{
		public static void PrintChildren(this Transform t)
		{
			for (int i = 0; i < t.childCount; i++)
			{
				t.GetChild(i).name.Print();
			}
		}
	}

	public static class JSONArrayExt
	{
		public static Vector3 ToV3(this JSONArray ja)
		{
			Vector3 v = new Vector3();
			for (int i = 0; i < 3; i++)
				v[i] = ja[i].AsFloat;
			return v;
		}
		
		public static Vector4 ToV4(this JSONArray ja)
		{
			Vector4 v = new Vector4();
			for (int i = 0; i < 4; i++)
				v[i] = ja[i].AsFloat;
			return v;
		}
	}

	public static class ColliderExt
	{
		public static string GetBodyRegion(this Collider rb)
		{
			if (rb.name.Contains("Link")) return null;
			else if (rb.name.Contains("Control")) return null;

			else if (rb.name.Contains("Lip")) return "Lips";
			else if (rb.name.Contains("FaceHardLeft5") || rb.name.Contains("FaceHardRight5"))
			{
				return "Eyes";
				rb.name.Print();
			}
			else if (rb.name.Contains("head")) return "Head";
			else if (rb.name.Contains("chest")) return "Chest";
			else if (rb.name.Contains("pelvis")) return "Pelvis";
			else if (rb.name.Contains("rThigh")) return "rThigh";
			else if (rb.name.Contains("lThigh")) return "lThigh";
			else if (rb.name.Contains("rShin")) return "rShin";
			else if (rb.name.Contains("lShin")) return "lShin";
			else if (!rb.name.Contains("rShoe") && (rb.name.Contains("rFoot") || rb.name.Contains("rBigToe") ||
			                                        rb.name.Contains("rSmallToe") ||
			                                        rb.name.Contains("rToe"))) return "rFoot";
			else if (!rb.name.Contains("lShoe") && (rb.name.Contains("lFoot") || rb.name.Contains("lBigToe") ||
			                                        rb.name.Contains("lSmallToe") ||
			                                        rb.name.Contains("lToe"))) return "lFoot";
			else if (rb.name.Contains("PhysicsMeshJointleft glute")) return "lGlutes";
			else if (rb.name.Contains("PhysicsMeshJointright glute")) return "rGlutes";
			else if (rb.name.Contains("AutoColliderFemaleAutoColliderslPectoral")) return "lBreast";
			else if (rb.name.Contains("AutoColliderFemaleAutoCollidersrPectoral")) return "rBreast";
			else if (rb.name.Contains("PhysicsMeshJointleft")) return "lBreast";
			else if (rb.name.Contains("PhysicsMeshJointright")) return "rBreast";
			else if (rb.name.Contains("JointG")) return "Labia";
			else if (rb.name.Contains("PhysicsMeshJointgen")) return "Labia";
			else if (rb.name.Contains("PhysicsMeshJointlab")) return "Labia";
			else if (rb.name.Contains("JointA")) return "Anus";
			else if (rb.name.Contains("PhysicsMeshJointan")) return "Anus";
			else if (rb.name.Contains("PhysicsMeshJointbridge")) return "Anus";
			// else if(rb.name.Contains("PhysicsMeshJoint")) return "Lips";
			else if (rb.name.Contains("chest")) return "Chest";
			else if (rb.name.Contains("abdomen")) return "Abdomen";
			else if (rb.name.Contains("pelvis")) return "Pelvis";
			else if (rb.name.Contains("lForeArm") || rb.name.Contains("lShldr")) return "lArm";
			else if (rb.name.Contains("rForeArm") || rb.name.Contains("rShldr")) return "rArm";
			else if (rb.name.Contains("lHand") || rb.name.Contains("lCarpal") || rb.name.Contains("lThumb") ||
			         rb.name.Contains("lIndex") || rb.name.Contains("lMid") || rb.name.Contains("lRing") ||
			         rb.name.Contains("lPinky")) return "lHand";
			else if (rb.name.Contains("rHand") || rb.name.Contains("rCarpal") || rb.name.Contains("rThumb") ||
			         rb.name.Contains("rIndex") || rb.name.Contains("rMid") || rb.name.Contains("rRing") ||
			         rb.name.Contains("rPinky")) return "rHand";

			else
			{
				// (name+" is not a valid region.").Print();
				return null;
			}
		}
		
		public static Atom GetAtom(this Collider collider)
		{
			Atom atom = null;
			Transform parent = collider.attachedRigidbody.transform.parent;
			int i = 0;
			while (atom == null && i < 20)
			{
				atom = parent.gameObject.GetComponent<Atom>();
				parent = parent.parent;
				i += 1;
			}

			return atom;
		}
	}

	public static class RigidbodyExt
	{
		static Vector3 lastVelocity;

		static Vector3 lastAngularVelocity;
		// public static Actor GetActor(this Rigidbody rb, CheesyFX_OmniForce script){
		// 	Atom atom = null;
		// 	FreeControllerV3 ctrl = rb.GetComponent<FreeControllerV3>();
		// 	if (ctrl != null) atom = ctrl.containingAtom;
		// 	else{
		// 		ForceReceiver fr = rb.GetComponent<ForceReceiver>();
		// 		if (fr != null) atom = fr.containingAtom;
		// 	}
		// 	return script.actors.First(x => x.atom == atom);
		// }

		public static Atom GetAtom(this Rigidbody rb)
		{
			Atom atom = null;
			Transform parent = rb.transform.parent;
			int i = 0;
			while (atom == null && i < 20)
			{
				atom = parent.gameObject.GetComponent<Atom>();
				// if(atom != null) parent.name.Print();
				parent = parent.parent;
				i += 1;
			}

			return atom;
		}

		public static Vector3 GetAcceleration(this Rigidbody rb)
		{
			Vector3 acceleration = (rb.velocity - lastVelocity) / Time.deltaTime;
			lastVelocity = rb.velocity;
			return acceleration;
		}

		public static Vector3 GetAngularAcceleration(this Rigidbody rb)
		{
			Vector3 angularAcceleration = (rb.angularVelocity - lastAngularVelocity) / Time.deltaTime;
			lastAngularVelocity = rb.angularVelocity;
			return angularAcceleration;
		}

		public static float GetUpAngle(this Rigidbody rb)
		{
			return Vector3.Dot(rb.transform.up, Vector3.up);
		}

		public static string GetBodyRegion(this Rigidbody rb)
		{
			string name = rb.name;
			string output = null;
			if (name.Contains("Link")) output = null;
			else if (name.Contains("Control")) output = null;

			else if (name.Contains("Lip")) output = "Lips";
			else if (name.Contains("FaceHardLeft5") || name.Contains("FaceHardRight5"))
			{
				output = "Eyes";
			}
			else if (name.Contains("head")) output = "Head";
			else if (name.Contains("chest")) output = "Chest";
			else if (name.Contains("pelvis")) output = "Pelvis";
			else if (name.Contains("rThigh")) output = "rThigh";
			else if (name.Contains("lThigh")) output = "lThigh";
			else if (name.Contains("rShin")) output = "rShin";
			else if (name.Contains("lShin")) output = "lShin";
			else if (name.Contains("rFoot") || name.Contains("rBigToe") || name.Contains("rSmallToe") ||
			         name.Contains("rToe")) output = "rFoot";
			else if (name.Contains("lFoot") || name.Contains("lBigToe") || name.Contains("lSmallToe") ||
			         name.Contains("lToe")) output = "lFoot";
			else if (name.Contains("PhysicsMeshJointleft glute")) output = "lGlutes";
			else if (name.Contains("PhysicsMeshJointright glute")) output = "rGlutes";
			else if (name.Contains("AutoColliderFemaleAutoColliderslPectoral")) output = "lBreast";
			else if (name.Contains("AutoColliderFemaleAutoCollidersrPectoral")) output = "rBreast";
			else if (name.Contains("PhysicsMeshJointleft")) output = "lBreast";
			else if (name.Contains("PhysicsMeshJointright")) output = "rBreast";
			else if (name.Contains("JointG")) output = "Labia";
			else if (name.Contains("PhysicsMeshJointgen")) output = "Labia";
			else if (name.Contains("PhysicsMeshJointlab")) output = "Labia";
			else if (name.Contains("JointA")) output = "Anus";
			else if (name.Contains("PhysicsMeshJointan")) output = "Anus";
			else if (name.Contains("PhysicsMeshJointbridge")) output = "Anus";
			// else if(name.Contains("PhysicsMeshJoint")) output = "Lips";
			else if (name.Contains("chest")) output = "Chest";
			else if (name.Contains("abdomen")) output = "Abdomen";
			else if (name.Contains("pelvis")) output = "Pelvis";
			else if (name.Contains("lForeArm") || name.Contains("lShldr")) output = "lArm";
			else if (name.Contains("rForeArm") || name.Contains("rShldr")) output = "rArm";
			else if (name.Contains("lHand") || name.Contains("lCarpal") || name.Contains("lThumb") ||
			         name.Contains("lIndex") || name.Contains("lMid") || name.Contains("lRing") ||
			         name.Contains("lPinky")) output = "lHand";
			else if (name.Contains("rHand") || name.Contains("rCarpal") || name.Contains("rThumb") ||
			         name.Contains("rIndex") || name.Contains("rMid") || name.Contains("rRing") ||
			         name.Contains("rPinky")) output = "rHand";
			else if (name.Contains("Gen")) output = "Penis";
			else if (name == "Testes") output = "Testes";
			else if (name == "sack") output = "DildoSack";
			else if (name == "b1" || name == "b2" || name == "b3") output = "DildoPenis";
			else if (name == "object") output = "Object";

			return output;
		}
		
	}


	public static class ObjectExt
	{
		static float timer = 0f;

		public static void Print(this object f)
		{
			SuperController.LogMessage(f.ToString());
		}

		public static void PrintEvery(this object f, float interval)
		{
			timer -= Time.fixedDeltaTime;
			if (timer < 0f)
			{
				timer = interval;
				SuperController.LogMessage(f.ToString());
			}
		}

		public static Collider GetColliderFromCol(this object obj)
		{
			Collider collider = null;
			if (obj is Collider) collider = (Collider)obj;
			if (obj is Collision) collider = ((Collision)obj).collider;
			else SuperController.LogError("Object has no Collider.");
			return collider;
		}

		public static Rigidbody GetRigidbodyFromCol(this object obj)
		{
			Rigidbody rigidbody = null;
			if (obj is Collider) rigidbody = ((Collider)obj).attachedRigidbody;
			if (obj is Collision) rigidbody = ((Collision)obj).rigidbody;
			// else SuperController.LogError("Object has no rigidbody.");
			else obj.GetType().Print();
			return rigidbody;
		}

		public static bool NullCheck(this object obj)
		{
			bool val = obj == null;
			SuperController.LogMessage($"{val}");
			return val;
		}
	}

	public static class IntExt
	{
		static float timer = 0f;

		public static void Print(this int f)
		{
			SuperController.LogMessage(f.ToString());
		}

		public static void PrintEvery(this int f, float interval)
		{
			timer -= Time.fixedDeltaTime;
			if (timer < 0f)
			{
				timer = interval;
				SuperController.LogMessage(f.ToString());
			}
		}
	}

	public static class FloatExt
	{
		static float timer = 0f;

		public static void Print(this float f)
		{
			SuperController.LogMessage(f.ToString());
		}

		public static void PrintEvery(this float f, float interval)
		{
			timer -= Time.fixedDeltaTime;
			if (timer < 0f)
			{
				timer = interval;
				SuperController.LogMessage(f.ToString());
			}
		}
	}

	public static class BoolExt
	{
		static float timer = 0f;

		public static void Print(this bool f)
		{
			SuperController.LogMessage(f.ToString());
		}

		public static void PrintEvery(this bool f, float interval)
		{
			timer -= Time.fixedDeltaTime;
			if (timer < 0f)
			{
				timer = interval;
				SuperController.LogMessage(f.ToString());
			}
		}
	}

	public static class ArrayExt
	{
		static float timer = 0f;

		public static void Print(this object[] f)
		{
			string[] sArray = f.Select(x => x.ToString()).ToArray();
			SuperController.LogMessage("[" + string.Join(", ", sArray) + "]");
		}

		// public static void PrintEvery(this bool f, float interval){
		// 	timer -= Time.fixedDeltaTime;
		// 	if(timer < 0f){
		// 		timer = interval;
		// 		SuperController.LogMessage(f.ToString());
		// 	}
		// }
	}

	public static class StringExt
	{
		static float timer = 0f;

		public static void Print(this string s)
		{
			SuperController.LogMessage(s);
		}

		public static void PrintEvery(this string s, float interval)
		{
			timer -= Time.fixedDeltaTime;
			if (timer < 0f)
			{
				timer = interval;
				SuperController.LogMessage(s);
			}
		}
	}

	public static class IEumeratorExt
	{
		public static IEnumerator Start(this IEnumerator ienum)
		{
			if (ienum != null) SuperController.singleton.StartCoroutine(ienum);
			return ienum;
		}

		public static void Stop(this IEnumerator ienum)
		{
			if (ienum != null) SuperController.singleton.StopCoroutine(ienum);
		}
	}

	public static class GameObjectExt
	{
		public static Atom GetAtom(this GameObject go)
		{
			ForceReceiver forceReceiver = go.GetComponentInParent<ForceReceiver>();
			Atom atom = forceReceiver?.containingAtom;
			if (atom == null)
			{
				atom = go.GetComponentInParent<Atom>();
			}

			return atom;
		}
	}

	public static class JSONStorableStringChooserExt
	{
		public static void Sort(this JSONStorableStringChooser sc)
		{
			List<string> choices = new List<string>(sc.choices);
			choices.Sort();
			sc.choices = choices;
		}
	}

	public static class AtomExt
	{
		public static JSONStorable GetStorable(this Atom atom, string name, bool exactMatch = false)
		{
			string storableId;
			if (!exactMatch)
			{
				storableId = atom.GetStorableIDs().FirstOrDefault(x => x.Contains(name));
				if (storableId != null)
				{
					return atom.GetStorableByID(storableId);
				}
				else return null;
			}
			// else storableId = atom.GetStorableIDs().FirstOrDefault(x => x == name);
			else return atom.GetStorableByID(name);
		}

		public static void PrintStorablesOfType<T>(this Atom atom)
		{
			foreach (var name in atom.GetStorableIDs())
			{
				if (atom.GetStorableByID(name) is T) name.Print();
			}
		}

		public static void PrintStorables(this Atom atom, string filter = null)
		{
			foreach (var name in atom.GetStorableIDs())
			{
				if (filter == null || name.ToLower().Contains(filter.ToLower()))
				{
					(name + ": " + atom.GetStorableByID(name).GetType()).Print();
				}
			}
		}

		public static bool IsVisible(this Atom atom)
		{
			if (atom.name == "CoreControl" || atom.name == "[CameraRig]" || atom.name == "Chokpahi-Asset") return false;
			Renderer[] renderers = atom.GetComponentsInChildren<Renderer>();
			return renderers.Any(x => x.enabled == true);
		}

		public static float GetActivity(this Atom atom)
		{
			float activity = 0f;
			foreach (ForceReceiver fr in atom.forceReceivers)
			{
				Rigidbody rb = fr.GetComponent<Rigidbody>();
				activity += rb.mass * Math.Abs(Vector3.Dot(rb.GetAcceleration(), rb.velocity));
				Vector3 angularAcc = rb.GetAngularAcceleration();
				for (int i = 0; i < 3; i++)
				{
					activity += Math.Abs(angularAcc[i] * rb.inertiaTensor[i] * rb.angularVelocity[i]);
				}
			}

			return activity;
		}

		public static List<string> GetInterestRegions(this Atom atom)
		{
			return new List<string>
			{
				"Eyes",
				"Lips",
				"Titts",
				"Ass",
				"Anus",
				"Pussy",
				"Legs",
				"Feet",
				// "Chest",
				// "Pelvis",
				// "Arms",
			};
		}

		public static List<Rigidbody> GetBodyRegion(this Atom atom, string name)
		{
			List<Rigidbody> rigidbodies = atom.rigidbodies.ToList();
			List<Rigidbody> output;
			switch (name)
			{
				case "Head":
					output = rigidbodies.Where(x => x.name.Contains("head") && !x.name.Contains("Control")).ToList();
					break;
				case "Lips":
					output = rigidbodies.Where(x => x.name == "LipTrigger").ToList();
					break;
				case "Mouth":
					output = rigidbodies.Where(x => x.name == "MouthTrigger").ToList();
					break;
				case "Throat":
					output = rigidbodies.Where(x => x.name == "ThroatTrigger").ToList();
					break;
				case "Chest":
					output = rigidbodies.Where(x => x.name.Contains("chest") && !x.name.Contains("Control")).ToList();
					break;
				case "Pelvis":
					output = rigidbodies.Where(x => x.name.Contains("pelvis") && !x.name.Contains("Control")).ToList();
					break;
				case "Abdomen":
					output = rigidbodies.Where(x => x.name.Contains("abdomen")).ToList();
					break;
				case "lBreast":
					output = rigidbodies.Where(x => x.name.Contains("lPectoral")).ToList();
					break;
				case "rBreast":
					output = rigidbodies.Where(x => x.name.Contains("rPectoral")).ToList();
					break;
				case "lNipple":
					output = rigidbodies.Where(x => x.name.Contains("lNippleTrigger")).ToList();
					break;
				case "rNipple":
					output = rigidbodies.Where(x => x.name.Contains("rNippleTrigger")).ToList();
					break;
				case "lGlutes":
					output = rigidbodies.Where(x => x.name.Contains("LGlute")).ToList();
					break;
				case "rGlutes":
					output = rigidbodies.Where(x => x.name.Contains("RGlute")).ToList();
					break;
				case "Labia":
					output = rigidbodies.Where(x => x.name == "LabiaTrigger").ToList();
					break;
				case "Vagina":
					output = rigidbodies.Where(x => x.name == "VaginaTrigger").ToList();
					break;
				case "DeepVagina":
					output = rigidbodies.Where(x => x.name == "DeepVaginaTrigger").ToList();
					break;
				case "DeeperVagina":
					output = rigidbodies.Where(x => x.name == "DeeperVaginaTrigger").ToList();
					break;
				case "Anus":
					output = rigidbodies.Where(x => x.name.Contains("_JointA")).ToList();
					break;
				case "lThigh":
					output = rigidbodies.Where(x => x.name.Contains("lThigh") && !x.name.Contains("Control")).ToList();
					break;
				case "rThigh":
					output = rigidbodies.Where(x => x.name.Contains("rThigh") && !x.name.Contains("Control")).ToList();
					break;
				case "lShin":
					output = rigidbodies.Where(x => x.name.Contains("lShin")).ToList();
					break;
				case "rShin":
					output = rigidbodies.Where(x => x.name.Contains("rShin")).ToList();
					break;
				case "lFoot":
					output = rigidbodies.Where(x => x.name.Contains("lFoot")).ToList();
					break;
				case "lBigToe":
					output = rigidbodies.Where(x => x.name.Contains("lBigToe")).ToList();
					break;
				case "lSmallToe1":
					output = rigidbodies.Where(x => x.name.Contains("lSmallToe1")).ToList();
					break;
				case "lSmallToe2":
					output = rigidbodies.Where(x => x.name.Contains("lSmallToe2")).ToList();
					break;
				case "lSmallToe3":
					output = rigidbodies.Where(x => x.name.Contains("lSmallToe3")).ToList();
					break;
				case "lSmallToe4":
					output = rigidbodies.Where(x => x.name.Contains("lSmallToe4")).ToList();
					break;
				case "rFoot":
					output = rigidbodies.Where(x => x.name.Contains("rFoot")).ToList();
					break;
				case "rBigToe":
					output = rigidbodies.Where(x => x.name.Contains("rBigToe")).ToList();
					break;
				case "rSmallToe1":
					output = rigidbodies.Where(x => x.name.Contains("rSmallToe1")).ToList();
					break;
				case "rSmallToe2":
					output = rigidbodies.Where(x => x.name.Contains("rSmallToe2")).ToList();
					break;
				case "rSmallToe3":
					output = rigidbodies.Where(x => x.name.Contains("rSmallToe3")).ToList();
					break;
				case "rSmallToe4":
					output = rigidbodies.Where(x => x.name.Contains("rSmallToe4")).ToList();
					break;
				// if(name == "lToes") output = rigidbodies.Where(x => x.name.Contains("lBigToe") && x.name.Contains("lIndex") && x.name.Contains("lMid") && x.name.Contains("lRing") && x.name.Contains("lPinky")).ToList();
				// if(name == "rFingers") output = rigidbodies.Where(x => x.name.Contains("rThumb") && x.name.Contains("rIndex") && x.name.Contains("rMid") && x.name.Contains("rRing") && x.name.Contains("rPinky")).ToList();
				case "lShoulder":
					output = rigidbodies.Where(x => x.name == "lShldr").ToList();
					break;
				case "rShoulder":
					output = rigidbodies.Where(x => x.name == "rShldr").ToList();
					break;
				case "lForeArm":
					output = rigidbodies.Where(x => x.name == "lForeArm").ToList();
					break;
				case "rForeArm":
					output = rigidbodies.Where(x => x.name == "rForeArm").ToList();
					break;
				case "lPalm":
					output = rigidbodies.Where(x => x.name.Contains("lHand") || x.name.Contains("lCarpal")).ToList();
					break;
				case "lThumb":
					output = rigidbodies.Where(x => x.name.Contains("lThumb")).ToList();
					break;
				case "lIndex":
					output = rigidbodies.Where(x => x.name.Contains("lIndex")).ToList();
					break;
				case "lMid":
					output = rigidbodies.Where(x => x.name.Contains("lMid")).ToList();
					break;
				case "lRing":
					output = rigidbodies.Where(x => x.name.Contains("lRing")).ToList();
					break;
				case "lPinky":
					output = rigidbodies.Where(x => x.name.Contains("lPinky")).ToList();
					break;
				case "rPalm":
					output = rigidbodies.Where(x => x.name.Contains("rHand") || x.name.Contains("rCarpal")).ToList();
					break;
				case "rThumb":
					output = rigidbodies.Where(x => x.name.Contains("rThumb")).ToList();
					break;
				case "rIndex":
					output = rigidbodies.Where(x => x.name.Contains("rIndex")).ToList();
					break;
				case "rMid":
					output = rigidbodies.Where(x => x.name.Contains("rMid")).ToList();
					break;
				case "rRing":
					output = rigidbodies.Where(x => x.name.Contains("rRing")).ToList();
					break;
				case "rPinky":
					output = rigidbodies.Where(x => x.name.Contains("rPinky")).ToList();
					break;
				// if(name == "lHand") output = rigidbodies.Where(x => x.name.Contains("lHand") || x.name.Contains("lCarpal") || x.name.Contains("lThumb") || x.name.Contains("lIndex") || x.name.Contains("lMid") || x.name.Contains("lRing") || x.name.Contains("lPinky")).ToList();
				// if(name == "rHand") output = rigidbodies.Where(x => x.name.Contains("rHand") || x.name.Contains("rCarpal") || x.name.Contains("rThumb") || x.name.Contains("rIndex") || x.name.Contains("rMid") || x.name.Contains("rRing") || x.name.Contains("rPinky")).ToList();
				default:
					(name + " is not a valid region.").Print();
					output = new List<Rigidbody>();
					break;
			}

			return output;
		}
	}

	public static class UIDynamicExt
	{
		public static void ForceHeight(this UIDynamic uIDynamic, float height)
		{
			LayoutElement component = uIDynamic.GetComponent<LayoutElement>();
			component.minHeight = height;
			component.preferredHeight = height;
		}
	}

	public static class JSONStorableParamExt
	{
		public static UIDynamic CreateUI(this JSONStorableParam jsonParam, List<object> UIElements = null,
			bool rightSide = false, int chooserType = 0, MVRScript script = null)
		{
			if (script == null) script = UIManager.script;
			UIDynamic uiDynamic = null;

			if (jsonParam is JSONStorableBool)
				uiDynamic = script.CreateToggle((JSONStorableBool)jsonParam, rightSide);
			else if (jsonParam is JSONStorableFloat)
				uiDynamic = script.CreateSlider((JSONStorableFloat)jsonParam, rightSide);
			else if (jsonParam is JSONStorableStringChooser)
			{
				if (chooserType == 0)
					uiDynamic = script.CreatePopup((JSONStorableStringChooser)jsonParam, rightSide);
				else if (chooserType == 1)
					uiDynamic = script.CreateScrollablePopup((JSONStorableStringChooser)jsonParam, rightSide);
				else if (chooserType == 2)
					uiDynamic = script.CreateFilterablePopup((JSONStorableStringChooser)jsonParam, rightSide);
			}
			else if (jsonParam is JSONStorableString)
			{
				JSONStorableString storableString = jsonParam as JSONStorableString;
				uiDynamic = Utils.SetupTextInput(script, storableString.name, storableString, rightSide);
			}
			else return null;

			if (UIElements != null) UIElements.Add(uiDynamic);
			return uiDynamic;
		}

		public static void Store(this JSONStorableParam jsonParam, JSONClass jc)
		{
			if (jsonParam is JSONStorableBool)
			{
				JSONStorableBool jParam = (JSONStorableBool)jsonParam;
				if (jParam.val != jParam.defaultVal) jc[jParam.name].AsBool = jParam.val;
			}
			else if (jsonParam is JSONStorableFloat)
			{
				JSONStorableFloat jParam = (JSONStorableFloat)jsonParam;
				if (jParam.val != jParam.defaultVal) jc[jParam.name].AsFloat = jParam.val;
			}
			else if (jsonParam is JSONStorableString)
			{
				JSONStorableString jParam = (JSONStorableString)jsonParam;
				if (jParam.val != jParam.defaultVal) jc[jParam.name] = jParam.val;
			}
			else if (jsonParam is JSONStorableStringChooser)
			{
				JSONStorableStringChooser jParam = (JSONStorableStringChooser)jsonParam;
				if (jParam.val != jParam.defaultVal) jc[jParam.name] = jParam.val;
			}
			else if (jsonParam is JSONStorableColor)
			{
				JSONStorableColor jParam = (JSONStorableColor)jsonParam;
				JSONArray ja = new JSONArray();
				ja[0].AsFloat = jParam.val.H;
				ja[1].AsFloat = jParam.val.S;
				ja[2].AsFloat = jParam.val.V;
				jc[jParam.name] = ja;
			}
		}

		public static void Load(this JSONStorableParam jsonParam, JSONClass jc)
		{
			if (jsonParam is JSONStorableBool)
			{
				JSONStorableBool jParam = (JSONStorableBool)jsonParam;
				if (jc.HasKey(jParam.name)) jParam.val = jc[jParam.name].AsBool;
			}
			else if (jsonParam is JSONStorableFloat)
			{
				JSONStorableFloat jParam = (JSONStorableFloat)jsonParam;
				if (jc.HasKey(jParam.name)) jParam.val = jc[jParam.name].AsFloat;
			}
			else if (jsonParam is JSONStorableString)
			{
				JSONStorableString jParam = (JSONStorableString)jsonParam;
				if (jc.HasKey(jParam.name)) jParam.val = jc[jParam.name];
			}
			else if (jsonParam is JSONStorableStringChooser)
			{
				JSONStorableStringChooser jParam = (JSONStorableStringChooser)jsonParam;
				if (jc.HasKey(jParam.name)) jParam.val = jc[jParam.name];
			}
			else if (jsonParam is JSONStorableColor)
			{
				JSONStorableColor jParam = (JSONStorableColor)jsonParam;
				if (jc.HasKey(jParam.name))
				{
					HSVColor loadedColor = new HSVColor
					{
						H = jc[jParam.name][0].AsFloat,
						S = jc[jParam.name][1].AsFloat,
						V = jc[jParam.name][2].AsFloat
					};
					jParam.val = loadedColor;
				}
			}
		}

		public static void Register(this JSONStorableParam jsonParam)
		{
			if (jsonParam is JSONStorableBool)
			{
				JSONStorableBool jParam = jsonParam as JSONStorableBool;
				UIManager.script.RegisterBool(jParam);
			}
			else if (jsonParam is JSONStorableFloat)
			{
				JSONStorableFloat jParam = jsonParam as JSONStorableFloat;
				UIManager.script.RegisterFloat(jParam);
			}
			else if (jsonParam is JSONStorableString)
			{
				JSONStorableString jParam = jsonParam as JSONStorableString;
				UIManager.script.RegisterString(jParam);
			}
			else if (jsonParam is JSONStorableStringChooser)
			{
				JSONStorableStringChooser jParam = jsonParam as JSONStorableStringChooser;
				UIManager.script.RegisterStringChooser(jParam);
			}
		}
	}

}

