using System;
using UnityEngine;

namespace CheesyFX
{
    public class RBDynamics : MonoBehaviour
    {
        public Rigidbody rigidbody;
        public Vector3 velocity;
        public Vector3 angularVelocity;
        private Vector3 _lastVelocity;
        private Vector3 _lastAngularVel;

        private Vector3 _acceleration;
        private bool _accelerationSet;
        
        private Vector3 _angularAcceleration;
        private bool _angularAccelerationSet;

        private float _activity;
        private bool _activitySet;

        public int uses;


        public void FixedUpdate()
        {
            _lastVelocity = velocity;
            _lastAngularVel = angularVelocity;
            velocity = rigidbody.velocity;
            angularVelocity = rigidbody.angularVelocity;
            _accelerationSet = false;
            _angularAccelerationSet = false;
            _activitySet = false;
        }
        public Vector3 acceleration
        {
            get
            {
                if (_accelerationSet) return _acceleration;
                _acceleration = (velocity - _lastVelocity) / Time.fixedDeltaTime;
                _accelerationSet = true;
                return _acceleration;
            }
        }

        public Vector3 angularAcceleration
        {
            get
            {
                if (_angularAccelerationSet) return _angularAcceleration;
                // var angularVelocity = angularVelocity;
                _angularAcceleration = (angularVelocity - _lastAngularVel) / Time.fixedDeltaTime;
                _angularAccelerationSet = true;
                return _angularAcceleration;
            }
        }
        public float activity
        {
            get
            {
                if (_activitySet) return _activity;
                float act = rigidbody.mass * Math.Abs(Vector3.Dot(acceleration, velocity));
                for (int i = 0; i < 3; i++)
                {
                    act += Mathf.Abs(angularAcceleration[i] * rigidbody.inertiaTensor[i] * angularVelocity[i]);
                }
                _activity = act;
                _activitySet = true;
                return act;
            }
        }
    }

    public static class Extensions
    {
        public static RBDynamics AddRbDynamic(this Rigidbody rb)
        {
            var rbDynamic = rb.gameObject.GetComponent<RBDynamics>();
            if (rbDynamic == null) rbDynamic = rb.gameObject.AddComponent<RBDynamics>();
            rbDynamic.rigidbody = rb;
            rbDynamic.uses ++;
            return rbDynamic;
        }
    }
}